/* tarech.cc       letzte Aenderung: 6.11.2010
Selbstbau-Taschenrechner
*/
#define VERSION "0.03"
/*
 Prozessor: Atmega32, 16MHz Quarz
 Schaltung: tarech.png
   die 3 LEDs auf PB0...PB2 sind noch unbenutzt (nicht angeschlossen)
   Stecker fuer TxD (serielle Ausgabe) noch nicht angeschlossen.
	    Transistor Q5 und SV1 nicht angeschlossen (anderes Projekt)
	    S1 noch unbenutzt
   Grafik-LCD: PC0...PC7, PD2-PD7  (noch nicht angeschlossen)
   Text-LCD: 4-Zeilen LCD an PortA (provisorisch statt Grafik-LCD)
   Tastatur: PD0 (RxD)  im Schema fehlt noch der +5V-Pin

 Autor: Rolf Pfister
 Copyright: Freeware
 History:
 19.9.2010	Erstellung, Uhr von Lichtwecker uebernommen (vereinfacht)
 6.11.          neue Tastenbelegung

*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
//#include <stdlib.h>
//#include <math.h>
#include <string.h>
#include <ctype.h>
#include "ulong.h"
#include "fliesszahlklasse.cc"

#define F_CPU 16000000UL
#define UARTVARIANTE //Methode fuer Tastatur-Anschluss
//#define GRAFIK_LCD //Wenn Grafik-LCD angeschlossen ist
#define TEXT_LCD   //Sonst mit Text-LCD

/**** Vordeklarationen ****/
void tastaturcheck();
char tastaturabfrage();
void glcd_printzeile(int zeile,const char *cstr,
		     int p1=0,int p2=0,int p3=0,int p4=0);
void tarech_init();
void tarech_tastenreaktion(uchar taste);
void hexprintf(char *str,const char *cstr,double x);

/**** Zeitbehandlung ****/
static volatile uint millisec=0;
static volatile uchar sec=0,min=02,stunden=19;
static volatile uchar tage=26,monat=10; //aktuelles Datum
static volatile uint jahr=2010;
static char monatstage[13]={0,31,28,31,30,31,30,31,31,30,31,30,31};
static volatile char wochentag=0; //0=Montag ... 6=Sonntag
static volatile char sommerzeit=1; //0=Winterzeit 1=Sommerzeit

void schaltjahrtest()
{
 if(jahr%4==0 && jahr%100!=0 || jahr%400==0) monatstage[2]=29;
 else monatstage[2]=28;
}
void wochentagberechnen()
{
 int j=jahr-2001,t,m;
 t=j*365+j/4-j/100+j/400+tage+6;
 for(m=1;m<monat;m++) t+=monatstage[m];
 wochentag=t%7;
}
//void variable_feiertage_berechnen();
void schaltjahr_und_wochentagberechnen()
{
 schaltjahrtest();
 wochentagberechnen();
 //variable_feiertage_berechnen();
}

/** Feiertage: ** /
const char NFEST=6;//Anzahl feste Feiertage
const char festfeiertag[NFEST]={1, 2, 1, 1, 25, 26};
const char festfeiermon[NFEST]={1, 1, 5, 8, 12, 12};
const char NVARI=4;//Anzahl von Ostern abhaengige Feiertage
static char varifeiertag[NVARI];
static char varifeiermon[NVARI];
void variable_feiertage_berechnen()
{
 int a,b,c,k,p,q,M,N;
 char d,e,tag,i;
 const char mon=3;
 //Osterformel nach Gauss:
 a=jahr%19;
 b=jahr%4;
 c=jahr%7;
 k=jahr/100;
 p=(8*k+13)/25;
 q=k/4;
 M=(15+k-p-q)%30;
 N=(4+k-q)%7;
 d=(19*a+M)%30;
 e=(2*b+4*c+6*d+N)%7;
 tag=22+d+e;
 varifeiertag[0]=tag+1; varifeiermon[0]=mon; //Ostermontag
 varifeiertag[1]=tag-2; varifeiermon[1]=mon; //Karfreitag
 varifeiertag[2]=tag+39; varifeiermon[2]=mon; //Auffahrt
 varifeiertag[3]=tag+50; varifeiermon[3]=mon; //Pfingstmontag
 for(i=0;i<NVARI;i++)
   while(varifeiertag[i]>monatstage[varifeiermon[i]])
    {varifeiertag[i] -= monatstage[varifeiermon[i]]; varifeiermon[i]++;}
}

bool istfeiertag()
{
 char i;
 for(i=0;i<NFEST;i++)
   if(tage==festfeiertag[i] && monat==festfeiermon[i]) return true;
 for(i=0;i<NVARI;i++)
   if(tage==varifeiertag[i] && monat==varifeiermon[i]) return true;
 return false;
}
/ **/

// Quarz moeglichst genau ausmessen und hier eintragen
// in Hertz und Tausendstel Hertz:

//const ulong Hz=16000000,MilliHz=0; //exakt 16MHz
//Beispiel: 16MHz ist nach n Tagen z sec vorgegangen:
//nach n Tagen sind also 16000000*(3600*24*n+z) Takte vergangen
//Frequenz ist also:
// x = (3600*24*n+z)*16000000 / (3600*24*n)
//   = 16000000 + z*16000000/(3600*24*n)
const ulong Hz=16002000,MilliHz=0;

const uint Rest1 = Hz%1000;//jede Sekunde zu korrigierender Fehler
//const uint Rest2 = MilliHz*6/100;//jede Minute
//const uint Rest3 = ((MilliHz*6-Rest2*100)*60+50)/100;//jede Stunde
const uint NormalerCompwert=Hz/1000-1;
//der Compiler sollte bis hier noch keinen Code erzeugt haben.
//(der Quarz ist wahrscheinlich zu wenig stabil als dass Rest2 und Rest3
//eine Rolle spielen wuerden. Also MilliHz vorlaeufig auf 0 lassen.)

//die Namen SIG_* sind in /usr/avr/include/avr/iom32.h zu finden
ISR(SIG_OUTPUT_COMPARE1A)
{
 uchar mehr;
 uint ms=millisec;
 uint h;//fuer LEDs
 if(++ms>=1000-Rest1) mehr=1; else mehr=0;
 sei();
 if(ms==1000)
  {ms=0;
   //if(++sec>=60-Rest2) mehr=1; else mehr=0;
   mehr=0;//einfacher wenn Rest2 und Rest3 sowieso ==0
   if(++sec==60)
    {sec=0;
     //if(++min>=60-Rest3) mehr=1; else mehr=0;
     if(++min==60)
      {min=0; mehr=0;
       if(++stunden==24)
	{stunden=0; ++tage; //neuertag=1;
	 if(++wochentag==7) wochentag=0; 
	 if(tage>monatstage[monat])
	  {tage=1;
	   if(++monat==13) //Jahreswechsel
	     {monat=1; ++jahr; sei(); schaltjahrtest();
	      //variable_feiertage_berechnen();
	     }
	  }
	}
       else //Test auf Sommerzeit-Umschaltungen
	{if(sommerzeit==0)
	  {if(stunden==2 && monat==3 && wochentag==6 && tage>(31-7))
	    {sommerzeit=1; stunden++;} //am letzten Sonntag im Maerz
	  }
	 else //if(sommerzeit==1)
	  {if(stunden==3 && monat==10 && wochentag==6 && tage>(31-7))
	    {sommerzeit=0; --stunden;} //am letzten Sonntag im Oktober
	  }
	}
      }
    }
  }
 //else if((ms&0x0F)==1) {}//etwa alle 16ms
 //else if((ms&0xFF)==0xE7) {}//alle ca 250ms
 millisec=ms;
 OCR1A=NormalerCompwert+mehr;
 tastaturcheck();
}

//Bei einem unerwarteten Interrupt nichts machen:
//SIGNAL (__vector_default) {}
//Oder besser wirklich einfach nur ein reti machen:
void __attribute__ ((naked)) __vector_default(void) {__asm__ ("reti");}
//Ohne diese Funktion wird nach 0 gesprungen, was einem Reset entspricht.

void milliwait(int n) //n Millisekunden warten  999 >= n >= 1
{
 uint ms;
 //ms=millisec+n;  das macht der Compiler nicht ganz korrekt!
 cli(); ms=millisec; sei(); ms+=n;
 if(ms>=1000) ms-=1000;
 while(ms!=millisec || ms!=millisec) ;
}

/**** Routinen fuer LCD-Ansteuerung ****/
void milliwait(int n);
char microwait(int n) //etwa n Microsekunden warten
{
 char vx=0;
 while(n>0)
  {vx+=n; --n;}
 return vx;
}

#ifdef GRAFIK_LCD
#include "glcd.cc"
#else
#include "lcd4zeilen.cc"
inline void glcd_init() {lcd_init();}
inline void glcd_goto(uchar zeile,uchar spalte) {lcd_goto(zeile,spalte);}
inline void glcd_write(char c) {lcd_write(c);}
inline void glcd_write(const char *txt) {lcd_write(txt);}
#endif

void lcd_gowrite(char zeile,char spalte,const char* text)
{
 glcd_goto(zeile,spalte); glcd_write(text);
}
void lcd_gowrite20(char zeile,const char* text)
{
 char txt[22],*s,i;
 for(i=0,s=txt; *s = *text++;s++,i++) ;
 for(;i<20;i++) *s++ = ' ';//Zeile mit Leerstellen auffuellen
 *s=0;
 lcd_goto(zeile,1); glcd_write(txt);
}
void sprintf_d(char *str,const char *cstr,double x)
{
 long ix,dx;
 int ex=0;
 bool isneg;
 char vorz=' ';
 if(x!=0)
  {for(ex=0;!(x>=0.1 || x<= -0.1);ex-=3,x*=1e3)
    ;//kleine Zahlen fuer Darstellung der Form 1.23e-6 vorbereiten
   for(;!(x<1000 && x> -1000);ex+=3,x*=0.001)
    ;//grosse Zahlen fuer Darstellung der Form 1.23e6 vorbereiten
  }
 isneg = x<0;
 if(isneg) vorz='-';
 ix=(long)x;
 if(isneg) {dx = (long)((ix-x)*1e6+0.5); ix = -ix;}
 else      {dx = (long)((x-ix)*1e6+0.5);} //jeweils +0.5 ist fuer korrekte Rundung
 if(dx==1000000)                          //wenn beim Runden auf eine Ganze Zahl aufgerundet wurde
   {dx=0;  ix = isneg ? ix-1 : ix+1;}     //dann entsprechend den Ganzzahligen Anteil korrigieren
 if(ex==0)
   sprintf(str,"%c%6ld.%06ld ",vorz,ix,dx);
 else
   sprintf(str,"%c%6ld.%06lde%d ",vorz,ix,dx,ex);
}

static int hexmodus=0;

void lcd_gowrite_d(char zeile,double x)
{
 char text[22];
 if(hexmodus) hexprintf(text,"%13lf",x);
 else sprintf_d(text,"%lf",x);
 lcd_gowrite20(zeile,text);
}

//Globale Variablen:
const char MINSEK=0,UHRZEIT=1,MOTG=2,WOTAG=3,JJJ=4,WECK1=5,WECK2=6;
static char uhrmodus=UHRZEIT,stellen=0;

void zeitauflcdanzeigen()
{
 char txt[40];
 glcd_goto(4,1);
 sprintf(txt,"%02d:%02d:%02d ",stunden,min,sec);
 glcd_write(txt);
 sprintf(txt,"%02d.%02d.%d",tage,monat,jahr);
 glcd_write(txt);
}

#ifdef UARTVARIANTE
void uart_init();
#endif

int main(void)
{
 char i,c,pos=1,modus=0,altsec=0;
 const int TASCHENRECHNER=1;//Gueltige Werte fuer modus

 TCCR1B = (1<<WGM12)|(1<<CS10); //CTC-OCR1A-Modus, Prescaler=1
 OCR1A = NormalerCompwert;
 TCNT1 = 0; //Startwert des Timers
 TIMSK = 1<<OCIE1A; //Timer1 Interrupts bei Vergleichswert

 sei(); //Interrupt einschalten
 schaltjahr_und_wochentagberechnen();

 milliwait(50);
 glcd_init();
 lcd_gowrite(1,1,"Taschenrechner");
 lcd_gowrite(2,1,"Version "); glcd_write(VERSION);
 lcd_gowrite(3,1,"Testmodus");//test
 // glcd_goto(4,1); glcd_write("2.Test: 4.Zeile");//test
#ifdef UARTVARIANTE
 uart_init();
#endif

 while(1) //Hauptschlaufe
  {if(modus==0)
    {if(altsec!=sec) {zeitauflcdanzeigen(); altsec=sec;}
     //milliwait(25);
     if((c=tastaturabfrage())!=0) //test
      {glcd_goto(3,pos); //Tastatureingabe auf 3.Zeile darstellen
       glcd_write(c);
       if(++pos==21) pos=1;
       if(c=='a')
	{pos=1; modus=TASCHENRECHNER;
	 lcd_gowrite(3,1,"Taschenrechner-Modus");
	 lcd_gowrite(4,1,"0                   ");
	 tarech_init();
	}
      }
    }
   else if(modus==TASCHENRECHNER)
    {//milliwait(25);
     if((c=tastaturabfrage())!=0)
      {
       tarech_tastenreaktion(c);
      }
    }
  }//Ende Hauptschlaufe
 return 0;//wird nie erreicht
}

/*** Abfrage der Matrix-Tastatur (mit Tiny2313 aufgebaut) ***/
#ifdef UARTVARIANTE

#ifndef F_CPU
#warning "F_CPU war noch nicht definiert, wird nun nachgeholt mit 16000000"
#define F_CPU 16000000UL  // Systemtakt in Hz - Definition als unsigned long
                         // Ohne ergeben sich unten Fehler in der Berechnung
#endif
 
#define BAUD 8192UL      // Baudrate
 
// Berechnungen
#define UBRR_VAL ((F_CPU+BAUD*8)/(BAUD*16)-1)   // clever runden
#define BAUD_REAL (F_CPU/(16*(UBRR_VAL+1)))     // Reale Baudrate
#define BAUD_ERROR ((BAUD_REAL*1000)/BAUD) // Fehler in Promille, 1000=kein Fehler

#if ((BAUD_ERROR<990) || (BAUD_ERROR>1010))
  #error Systematischer Fehler der Baudrate groesser 1% und damit zu hoch! 
#endif

void uart_init()
{
 UBRRH = UBRR_VAL >> 8;
 UBRRL = UBRR_VAL & 0xFF;
 //UCSRB |= (1<<TXEN); // UART TX einschalten
 UCSRB |= (1<<RXEN); // UART RX einschalten
 UCSRC = (1<<URSEL)|(3<<UCSZ0); // Asynchron 8N1 Atmega32
}

/*
void uart_putc(unsigned char c)
{
 while (!(UCSRA & (1<<UDRE))) ;// warten bis Senden moeglich
 UDR = c;                      // sende Zeichen
}
char uart_getc(void)
{
 while (!(UCSRA & (1<<RXC))) ;// warten bis Zeichen verfuegbar
 return UDR; // Zeichen aus UDR an Aufrufer zurueckgeben
}
*/

static char uart_puffer[8],uart_i1=0,uart_i2=0;
void tastaturcheck()
{
 if(!(UCSRA & (1<<RXC))) return;// warten bis Zeichen verfuegbar
 if(((uart_i2+1)&0x07)==uart_i1) return;//Puffer voll, noch nichts tun
 uart_puffer[uart_i2++]=UDR;
 uart_i2 &= 0x07;
}
char tastaturabfrage()
{
 char c=0;
 if(uart_i1!=uart_i2)
  {c=uart_puffer[uart_i1++]; uart_i1&=0x07;}
 return c;
}

#else
/*
Datenuebertragung auf 1 Leitung:
-------------------------------

                | 1.Bit |  2. Bit  |   ...  9. Bit |    Rueckmeldung

-----+ Startbit +---+   +---+      +---            +----+   +--
     |  3 ms    |1ms|1ms|1ms| 2ms  |   ...         |    |1ms|
     |          |   |   |   |      |               |    |   |
     +----------+   +---+   +------+             --+    +---+

Das Startbit definiert die Laenge eines zu uebertragenden Bits und ist
3 Einheiten lang (z.B. 3ms). Eine 0 wird durch einen 2 Einheiten langen
Puls uebertragen, eine 1 durch einen 1 Einheit langen Puls.
(Eine 1 also ein kurzer negativer Puls, eine 0 ein langer Puls)
Wenn das erste Datenbit eine 1 ist wird ein Kommando fuer das LCD, sonst
ein darzustellendes Zeichen uebertragen. Die folgenden 8 Bit ergeben
ein Byte beginnend mit dem hoechstwertigen Bit.
Fuer eine schnelle Uebertragung muss am Ende die Datenrichtung umgekehrt
werden, und der LCD-Kontroller gibt einen 1 Einheiten langen Puls zurueck.
(Fuer langsame Uebertragung kann man darauf verzichten und einfach eine
genuegend lange Warteschlaufe einbauen.)
*/
static char taste_ready=0,taste_c=0;
void tastaturcheck()
{
 char d0;
 d0=PIND&1;
 if(d0==1) return;
 char i,j,ok=0;
 int startbit,kurzesbit,langesbit,sehrlangesbit,tmitte,t1,daten;
 for(j=0;ok==0 && j<3;j++)
  {for(startbit=0;(PIND&1)==0;startbit++)
     microwait(10);//Laenge des Startbits messen
   langesbit=startbit*2/3;
   kurzesbit=startbit/3;
   sehrlangesbit=langesbit+(kurzesbit+1)/2;
   tmitte=(langesbit+kurzesbit+1)/2;
   for(i=0,daten=0,ok=1;i<9;i++)
    {for(t1=0;(PIND&1)!=0;t1++)
      {microwait(10);//Low abwarten
       if(t1>sehrlangesbit) {ok=0; daten=0; break;} //Abbruch bei Fehler
      }
     if(ok==0) break;
     for(t1=0;(PIND&1)==0;t1++) microwait(10);//Bitlaenge messen
     daten <<= 1;
     if(t1<tmitte) daten++;
     else if(t1>sehrlangesbit) {ok=0; daten=0; break;} //Abbruch bei Fehler
    }
  }
 if((d0=daten&0xFF)!=0) {taste_ready=1; taste_c=d0;}
 return;
}
char tastaturabfrage()
{
 char c=0;
 if(taste_ready)
  {c=taste_c; taste_ready=0;}
 return c;
}
#endif

/*************** Uebernommen von vereinfachtem Vektar ****************/

//(3.9.08) Vermeidung von Konflikten mit anderweitig definiertem abs:
#ifdef abs
#undef abs
#endif
#define abs meinabs

typedef double Vektor3d;
//const double GRAD=atan(1.0)*4/180;
//const double PI=atan(1.0)*4;
const double PI=3.14159265358979323846;
const double GRAD=PI/180;

#define INDEX(s1,s2) s_index(s1,strlen(s1),s2,1)

/************************* Vordeklarationen ***************************/
int keyget(int* in,int *asci,ULONG* rawcode);

/**************************** Kleinkram *******************************/
inline bool ist(const char*t,const char*s) {return strcmp(t,s)==0;}
//inline int idfix(double x) {return x<0?int(x-0.5):int(x+0.5);}
inline double abs(double x) {return x<0?-x:x;}

//static int hexmodus=0; //vorgezogen wegen lcd_gowrite_d()

void hexprintf(char *str,const char *cstr,double x)
{
 int i,n,c,imax,imax2;
 long long mx;
 double x1,x2;
 if(*cstr=='%') sscanf(&cstr[1],"%d",&imax); else imax=13;
 if(x<0) {*str++ = '-'; x= -x;}
 mx=(long long)x;
 x1=x-mx;
 x2=mx/16.0;
 for(i=1;x2>=1.0;i++) x2/=16.0;
 if(i>imax) i=imax;
 imax2=imax-i;
 for(;--i>=0;)
   {x2*=16; n=int(x2);
    x2-=n;
    if(n>9) c='A'+n-10; else c='0'+n;
    *str++ = c;
   }
 if(imax2>1) {*str++ = '.'; imax2--;}
 for(i=1;i<imax2 && x1>0;i++)
   {x1*=16;
    n=int(x1);
    x1-=n;
    if(n>9) c='A'+n-10; else c='0'+n;
    *str++ = c;
   }
 for(;i<=imax2;i++) *str++ = ' ';
 *str=0;
}

void lcd_goprint(int8 zeile,char *str)
{
 char c, *s;
 lcd_goto(zeile,1);
 while(strlen(str) > 20)
   {
    c=str[20]; str[20]=0;
    lcd_write(str);
    str[20]=c;
    str = &str[20];
    if(++zeile > 4) zeile=1;
    lcd_goto(zeile,1);
   }
 lcd_write(str);
}

void wprintf(int zeile,const char *cs,const char *p1,const char *p2=NULL)
{
 char str[80];
 sprintf(str,cs,p1,p2);
 lcd_goprint(zeile,str);
}
void wprintf(int zeile,const char *cs,int p1,int p2=0)
{
 char str[80];
 sprintf(str,cs,p1,p2);
 lcd_goprint(zeile,str);
}

void minus_entfernen(char *s)
{
 char *t= &s[1];
 while(*s++ = *t++) ;
}
void minus_einfuegen(char *s)
{
 char c='-',c2;
 while(c)
   {c2= *s; *s++ = c; c=c2;}
 *s=0;
}

int s_index(const char *s1,int n1,const char *s2,int n2)
			/* Sucht den String s2 innerhalb von s1 und  */
{			/* gibt die Position zurueck (nicht gefunden = -1) */
 int i,c,j;
 const char *p1,*p2;
 if(n2<=0) return 0;	/* leerer String ist immer enthalten */
 for(i=0;;i++)
	{if(i>=n1) return -1; /* nicht gefunden */
	 c= *s1++;
	 if(c== *s2)
		{for(p1=s1,p2=s2,j=1; j<n2; j++)
			{c= *++p2;
			 if(*p1++!=c) break; /* noch nicht gefunden */
			}
		 if(j==n2) break; /* gefunden */
		}
	}
 return i;
}

/************************* Fehlermeldungen ****************************/
class Fehlermeldung
{
public:
 int errflag;
 Fehlermeldung() {errflag=0;}
 void error(const char *s,int p1=0,int p2=0);
 void error(const char *s,const char *p1);
 void say(const char *s,const char *p1=NULL,const char *p2=NULL);
 // void clear() {color(0); fillbox(x0,y0,x2,y2); wfont(1);}
 void clear();
};

void Fehlermeldung::error(const char *s,int p1,int p2)
{
 clear();
 wprintf(4,s,p1,p2);
 if(*s!=' ') errflag=1; else errflag=0;
}
void Fehlermeldung::error(const char *s,const char *p1)
{
 clear();
 wprintf(4,s,p1);
 if(*s!=' ') errflag=1; else errflag=0;
}
void Fehlermeldung::say(const char *s,const char *p1,const char *p2)
{
 clear();
 wprintf(4,s,p1,p2);
 errflag=0;
}
void Fehlermeldung::clear()
{
 wprintf(4,"                    ",0,0);
 errflag=0;
}

/************************ Globale Variablen ***************************/
const int TYP_SHIFT=1,TYP_GOLD=2,TYP_NORMAL=0,TYP_HEXTASTE=3;
const int RADIAN=0,DEGREE=1;

static int exitflag=0;
static Fehlermeldung meldung;
static Vektor3d regist[10];

/***************************** Klassen ********************************/
const int STAPELSIZE=4;

class Stapel
{
 double mem[STAPELSIZE];
 char imax;
public:
 Stapel(char n=STAPELSIZE);
 void draw();
 void push(Vektor3d a);
 Vektor3d pop();
 Vektor3d rol(bool);
 Vektor3d get(char j);
 void set(char j,Vektor3d a) {if(j>=imax) j=imax-1; mem[j]=a;}
};
Stapel::Stapel(char n)
{
 imax=n;
 //for(char i=0;i<n;i++) mem[i]=0;
 for(char i=0;i<n;i++) mem[i]=(double)i+0.1;//test
}
void Stapel::draw()
{
 for(char i=0;i<4;i++)
   lcd_gowrite_d(4-i,get(i));
}
Vektor3d Stapel::get(char j)
{
 if(j>=imax) j=imax-1;
 return mem[j];
}
void Stapel::push(Vektor3d a)
{
 for(char i=imax-1;i>0;i--) mem[i]=mem[i-1];
 mem[0]=a;
}
Vektor3d Stapel::pop()
{
 Vektor3d a=mem[0];
 for(char i=1;i<imax;i++) mem[i-1]=mem[i];
 return a;
}
Vektor3d Stapel::rol(bool rev)
{
 Vektor3d a;
 char i;
 if(rev)
   {for(a=mem[i=imax-1]; i>0; i--)  mem[i]=mem[i-1];
    mem[0]=a;
   }
 else
   {a=mem[0];
    for(i=1;i<imax;i++) mem[i-1]=mem[i];
    mem[imax-1]=a;
   }
 return a;
}

class Eingabefeld
{
 bool aktiv,reop;
 char str[40];
 int cursor,nx;
 Stapel *ps;
public:
 bool help;
 Eingabefeld() {aktiv=reop=help=0; nx=0;}
 void init(Stapel *p) {ps=p;}
 void open(int n) {nx=n; str[cursor=0]=0; aktiv=true; reop=0;}
 void reopen(int n);
 void write(int c);
 void write(const char *s) {int c;while((c= *s++)!=0) write(c);}
 double close();
 void draw();
 bool istaktiv() {return aktiv;}
};
void Eingabefeld::reopen(int n)
{
 double v;
 if(aktiv) close();
 nx=n;
 v=ps->get(0);
 sprintf_d(str,"%lg",v);
 cursor=strlen(str);
 if(ist(str,"0")) str[cursor=0]=0;
 aktiv=reop=true;
}
bool ishexdigit(int c)
{
 if(c>='A') return (c<='F' || (c>='a' && c<='f'));
 return isdigit(c);
}
bool isdigitx(int c)
{
 if(hexmodus) return ishexdigit(c);
 return isdigit(c);
}
int hexdigit(int c)
{
 if(c>='a') return c-'a'+10;
 if(c>='A') return c-'A'+10;
 return c-'0';
}
double pow(double x,int i)
{
 double z=1.0;
 if(i>0)  {do {z *= x;} while(--i!=0);}
 else if(i<0) {do {z /= x;} while(++i!=0);}
 return z;
}
void hexsscanf(char *str,const char *cstr,double* px)
{
 double x=0,y,ex,v=1;
 const char *s=str;
 int c;
 if(*s=='-') {v= -1; s++;}
 for(;(c= *s++)!=0 && ishexdigit(c);)
   x=x*16+hexdigit(c);
 if(c=='.')
   {for(y=1.0/16.0;(c= *s++)!=0 && ishexdigit(c);y/=16.0)
       x+=y*hexdigit(c);
   }
 if(c=='*' && (c= *s++)=='1' && (c= *s++)=='0' && (c= *s++)=='^')
   {int vorz=1;
    if(*s=='-') {vorz= -1; s++;}
    for(ex=0;(c= *s++)!=0 && ishexdigit(c);)
       ex=ex*16+hexdigit(c);
    x *= pow(16.0,ex*vorz);
   }
 *px=v*x;
}
void dezsscanf(char *str,const char *cstr,double* px)
{
 double x=0,y,ex,v=1;
 const char *s=str;
 int c;
 if(*s=='-') {v= -1; s++;}
 for(;(c= *s++)!=0 && isdigit(c);)
   x=x*10+(c-'0');
 if(c=='.')
   {for(y=1.0/10.0;(c= *s++)!=0 && isdigit(c);y/=10.0)
       x+=y*(c-'0');
   }
 if(c=='e' || c=='E' ||
    (c=='*' && (c= *s++)=='1' && (c= *s++)=='0' && (c= *s++)=='^'))
   {int vorz=1;
    if(*s=='-') {vorz= -1; s++;}
    for(ex=0;(c= *s++)!=0 && isdigit(c);)
       ex=ex*10+(c-'0');
    x *= pow(10.0,ex*vorz);
   }
 *px=v*x;
}
double Eingabefeld::close()
{
 double x;
 x=ps->get(0);
 if(reop)
   {reop=0;}
 else
   {if(*str==0) x=0;
    else if(hexmodus) hexsscanf(str,"%lf",&x);
    else dezsscanf(str,"%lf",&x);
    ps->set(0,x);
   }
 aktiv=0;
 return x;
}
void Eingabefeld::write(int c)
{
 static int emi=0;
 if(!aktiv) open(nx);
 reop=0;
 if(c=='\b')
   {if(cursor==0) return;//Feld leer
    str[--cursor]=0;
    return;
   }
 if(cursor>=sizeof(str)-1) return;//Feld voll
 if(c=='-')
   {int i;
    for(i=cursor;i>0 && str[--i]!='+';)
      {if(!hexmodus) {if(str[i]=='e' && (emi=1-emi)==1) {i++; break;}}
       else {if(str[i]=='^' && (emi=1-emi)==1) {i++; break;}}
      }
    if(str[i]=='+') str[i]='-';
    else if(str[i]=='-') {minus_entfernen(&str[i]); --cursor;}
    else {minus_einfuegen(&str[i]); ++cursor;}
    return;
   }
 if(hexmodus==0 && c=='e' && cursor==0) str[cursor++]='1';
 str[cursor++]=c;
 str[cursor]=0;
}
void Eingabefeld::draw()
{
 if(aktiv)
   {char text[22];
    sprintf(text,"%18s  ",str);
    lcd_gowrite(4,1,text);
   }
}

class Taste
{
 const char **text0,**textn,**textb,**textg,**textgb;
 int typ;
public:
 //Taste() {typ=TYP_NORMAL; text0=NULL;}
 void init(const char**t1,const char**t2,const char**t3,const char**t4)
        {text0=textn=t1; textb=t2; textg=t3; textgb=t4;
	 typ=TYP_NORMAL;
	}
 int press(int,Eingabefeld*,Stapel*);
 void shift(int i) {text0 = i==0?textn:(i==1?textb:(i==2?textg:textgb));}
};

//Die Tasten "Help", "Blue" und "Gold" muessen bei allen Tastenbelegungen
//an der gleichen Position sein, die andern Tasten koennen beliebig
//angeordnet werden.
//In aktueller Version muss "Gold" an 1.Stelle und "Blue" an 2.Stelle stehen.

/*
Tasten wie sie gesendet werden:
 {'a','b','c','d','e',
  'A','B','C','D','/',
  '7','8','9','E','*',
  '4','5','6','F','-',
  '1','2','3','N','+',
  '0','.','I','J','='}
 */
static const char *tastenname[]=
 {"Gold","Blue","Help","POP", "CORR",
  "A",   "B",   "C",   "D",   "/",
  "7",   "8",   "9",   "E",   "*",
  "4",   "5",   "6",   "F",   "-",
  "1",   "2",   "3",   "EXP", "+",
  "0",   ".",   "+/-", "x<>y","ENTR"
 };
static const char *tastenname_gold[]=
 {"Gold","Blue","Help","POP", "CORR",
  "Sin", "Cos", "Tan", "Deg", "/",
  "7",   "8",   "9",   "ln",  "*",
  "4",   "5",   "6",   "log", "-",
  "1",   "2",   "3",   "EXP", "+",
  "0",   ".",   "+/-", "x<>y","ENTR"
 };
static const char *tastenname_blue[]=
 {"Gold","Blue","Help","ROL^","ROLv",
  "Asin","Acos","Atan","Rad",  "/",
  "7",   "8",   "9",   "e^x",  "*",
  "Pi",  "5",   "6",   "10^x", "-",
  "Date","SQRT","x^2", "y^x",  "+",
  "Time"," 1/x", "HEX", " Dez",  "ENTR"
 };
static const char *tastenname_goldblue[]=
 {"Gold","Blue","Help","ROL^","ROLv",
  "Asin","Acos","Atan","Rad",  "/",
  "Kale","MG-R","9",   "e^x",  "*",
  "4",   "5",   "6",   "10^x", "-",
  "1",   "SQRT","x^2", "y^x",  "+",
  "0",  " 1/x", "HEX", " Dez",  "ENTR"
 };
static const char *tasten_hilfe[]=
 {
  /*
"ALFA","berechnet Winkel zwischen x und y",
  "RAD",
    "Umschaltung zwischen GRAD und RAD fuer Winkelberechnungen (180GRAD==Pi)",
  "Pi","Eulersche Kreiszahl 3.14159265...",
  " e","Basis des natuerlichen Logarithmus 2.71828...",
  "SQRT","Wurzel",
  " 1/x","Kehrwert",
  "SIN","sin(x)  Sinus",
  "COS","cos(x)  Cosinus",
  "TAN","tan(x)  Tangens",
  "ASIN","arcsin(x)  Umkehrfunktion zu Sinus",
  "ACOS","arccos(x)  Umkehrfunktion zu Cosinus",
  "ATAN","arctan(x)  Umkehrfunktion zu Tangens",
  "LN","ln(x)  Natuerlicher Logarithmus",
  "LOG","log(x)  ZehnerLogarithmus",
  "-->|","Tabulator",
  "|<--","Rueckwaerts-Tabulator",
  "POP","x wird geloescht, das oberste Stapelelement dubliziert",
  "ROLv","der Stapel wird rotiert",
  "ROL^","der Stapel wird nach oben rotiert",
  "ROL->","x.X --> x.Y --> x.Z  rotieren",
  "ROL<-","x.X <-- x.Y <-- x.Z  rotieren",
  "RCL","Register 0 wird gelesen",
  "STO","x wird in Register 0 gespeichert",
  "RCLx","Register x wird gelesen, x muss zwischen 0 und 9 sein",
  "STOx","y wird in Register x gespeichert, x muss zwischen 0 und 9 sein",
  "+/-","Vorzeichen wechseln",
  ".","Dezimalpunkt",
  "EHV","Einheitsvektor erzeugen, Richtung bleibt gleich, Betrag wird 1",
  "x^2","x quadrieren",
  "x<>y","x und y vertauschen",
  "|x|","Betrag von x",
  "-","y-x --> x",
  "Blue","Umschaltung auf andere Funktionen, meistens inverse Funktionen",
  "CORR","Zahleneingabe korrigieren",
  "Gold","Umschaltung auf weitere Funktionen",
  "EXP","Exponent eingeben",
  "x","Vektormultiplikation: y*x --> x",
  "*","einfache Multiplikation:  y.X*x.X y.Y*x.Y  y.Z*x.Z",
  "CPLX","mit Complexen Zahlen rechnen (geht noch nicht)",
  "|xxx|","Betrag von x auf XYZ kopiert",
  "|x00|","x=unveraendert y=0 z=0",
  "ZROT","Rotiere y um den Winkel |x| um die Z-Achse (im Uhrzeigersinn)",
  "-ZROT","Rotiere y um den Winkel |x| um die Z-Achse im Gegenuhrzeigersinn",
  "YROT","Rotiere y um den Winkel |x| um die Y-Achse",
  "XROT","Rotiere y um den Winkel |x| um die X-Achse",
  "yROT","rotiert z um die durch y gegebene Achse um den Winkel |x|",
  "yzROT","rotiert t um die durch z-y gegebene Achse um den Winkel |x| (noch fehlerhaft)",
  "-YROT","wie YROT aber im Gegenuhrzeigersinn",
  "-XROT","wie XROT aber im Gegenuhrzeigersinn",
  "-yROT","wie yROT aber im Gegenuhrzeigersinn",
  "-yzROT","wie yzROT aber im Gegenuhrzeigersinn",
  "POLAR","Umrechnung von Polarkoordinaten (alfa,phi,Betrag) in XYZ",
  "RALOP","Umrechnung in Polarkoordinaten, X=alfa Y=phi Z=Betrag",
  "HEX","Hexadezimal-Modus einschalten",
  " Dez","Dezimal-Modus einschalten (HEX aus)",
  */
  "Gold","Umschaltung1",//test
  "Blue","Umschaltung2",//test
  NULL,NULL
 };

class Tastenfeld
{
 int nreihen,nspalten;
 Taste tasten;
 //int blue;
public:
 Tastenfeld()
   {nreihen=6; nspalten=5;
    tasten.init(tastenname,tastenname_blue,tastenname_gold,tastenname_goldblue);
   }
 void press(int i,Eingabefeld *ef,Stapel *ps);
 void shift(char j) {tasten.shift(j);}
};
void Tastenfeld::press(int i,Eingabefeld *ef,Stapel *ps)
{
 tasten.press(i,ef,ps);
}

//int gettastennummer(const char *txt,int j=0)
int gettastennummer(const char *txt)
{
 int i,n=sizeof(tastenname)/sizeof(char*);
 const char **name=tastenname;
 /*
 switch(j)
   {case 0: name=tastenname; break;
    case 1: name=tastenname_blue; break;
    case 2: name=tastenname_gold; break;
    case 3: name=tastenname_goldblue;
   }
 if(strcmp(txt,"*")==0 && j&1) txt="x";
 */
 for(i=0;i<n;i++)
   {if(strcmp(txt,name[i])==0) return i;
   }
 return -1;
}

/*
int gettastennummer(int c,int j=0)
{
 char s[2]; s[0]=c; s[1]=0;
 return gettastennummer(s,j);
}
*/

static int winkelmodus=DEGREE; //oder RADIAN

double rad(double x)
{
 if(winkelmodus==DEGREE) x*=GRAD;
 return x;
}

double grad(double x)
{
 if(winkelmodus==DEGREE) x/=GRAD;
 return x;
}

class Vektar
{
 Stapel stapel;
 Eingabefeld eingabefeld;
 Tastenfeld tafeld;
public:
 //Vektar() {}
 void init();
 void press(int i);
 void shift(char j) {tafeld.shift(j);}
 bool eingabeaktiv() {return eingabefeld.istaktiv();}
};
void Vektar::init()
{
 eingabefeld.init(&stapel);
}
void Vektar::press(int i)
{
 tafeld.press(i,&eingabefeld,&stapel);
 stapel.draw();
 eingabefeld.draw();
}
static Vektar vektar;

void gethelp(const char *frage)
{
 const char **p,*s,*antw=NULL;
 for(p=tasten_hilfe; (s= *p++)!=NULL; p++)
   {if(strcmp(s,frage)==0) {antw = *p; break;}
   }
 if(antw==NULL) meldung.say("'%s' no help yet ",frage);
 else meldung.say("'%s' %s ",frage,antw);
}

double sqrt(double x)
{
 double result=1,r0=x;
 int i;
 for(i=0;i<100 && result!=r0;i++)
   {r0 = result;
    result = (r0+x/r0)/2;
   }
 return result;
}

/************************* Hauptprogramm ******************************/
int Taste::press(int j,Eingabefeld *ef,Stapel *stap)
{
 static int blue=0; // 1 = Taste Blue aktiv, 2 = Taste Gold aktiv, 3 = beide aktiv
 static int eing=0,nx=0;
 int c,c2;
 double x,y,z;
 //drawselect(ef->help);
 //c= *text;
 const char *text=text0[j];
 c= *text;
 //printf("press(Eingabefeld,Stapel) c=%d='%d'\n",c,c);//test
 if(ef->help) {gethelp(text); ef->help=0; return blue;}
 if(ist(text,"Blue")) return blue ^= 1;
 if(ist(text,"Gold")) return blue ^= 2;
 /*
 if(ist(text,"-->|") || ist(text,"|<--"))
   {if(eing==0) {stap->push(0); ef->open(nx); eing=1;}
    else if(c=='-') ef->reopen(nx=(nx==2)?0:nx+1);
    else  ef->reopen(nx=(nx==0)?2:nx-1);
    return blue;
   }
 */
 if(ist(text,"ENTR"))
   {if(eing) {ef->close(); eing=0;}
    else stap->push(stap->get(0));
    return blue;
   }
 if(eing==0)
   {if(isdigitx(c) || c=='.' || ist(text,"EXP"))
      {stap->push(0); ef->open(nx); eing=1; //Eingabemodus einschalten
      }
    else if(ist(text,"CORR"))
      {ef->reopen(nx); eing=1; //Eingabemodus einschalten
       return blue;
      }
   }
 if(eing==1)
   {if(isdigitx(c)||c=='.'||ist(text,"EXP")||ist(text,"+/-")||ist(text,"CORR"))
      {//Reaktion auf Tasten im Eingabemodus:
       if(ist(text,"+/-")) c='-';
       else if(ist(text,"EXP")) {c='e'; if(hexmodus){ef->write("*10");c='^';}}
       else if(ist(text,"CORR")) c='\b';
       ef->write(c);
       return blue;
      }
    else
      {ef->close(); eing=0;} //Eingabemodus beenden
   }
 if(ist(text,"+/-")) {x=stap->pop(); stap->push(-x);}
 else if(ist(text,"POP")) stap->pop();
 else if(ist(text,"ROLv")) stap->rol(0);
 else if(ist(text,"ROL^")) stap->rol(1);
 else if(ist(text,"+")) {x=stap->pop(); y=stap->pop(); stap->push(y+x);}
 else if(ist(text,"-")) {x=stap->pop(); y=stap->pop(); stap->push(y-x);}
 else if(ist(text,"/"))
   {x=stap->pop(); y=stap->pop();
    if(x==0) stap->push(1e31);
    else stap->push(y/x);
   }
 else if(ist(text,"*"))	{x=stap->pop(); y=stap->pop();stap->push(y*x);}
 /* provi. ohne Winkelfunktionen
 else if(ist(text,"SIN")) {x=stap->pop(); stap->push(sin(rad(x)));}
 else if(ist(text,"COS")) {x=stap->pop(); stap->push(cos(rad(x)));}
 else if(ist(text,"TAN")) {x=stap->pop(); stap->push(tan(rad(x)));}
 else if(ist(text,"ASIN")) {x=stap->pop(); stap->push(grad(asin(x)));}
 else if(ist(text,"ACOS")) {x=stap->pop(); stap->push(grad(acos(x)));}
 else if(ist(text,"ATAN")) {x=stap->pop(); stap->push(grad(atan(x)));}
 */
 else if(ist(text,"SQRT")) {x=stap->pop(); stap->push(sqrt(x));}
 else if(ist(text,"x^2")) {x=stap->pop(); stap->push(x*x);}
 else if(ist(text,"x<>y"))
	{x=stap->pop(); y=stap->pop(); stap->push(x); stap->push(y);}
 else if(ist(text,"|x|"))
	{x=stap->pop(); x=abs(x); stap->push(x);}
 else if(ist(text,"Pi")) {stap->push(PI);}
 /* provi. ohne Logarithmen
 else if(ist(text," e")) {stap->push(exp(1.0));}
 else if(ist(text,"LN")) {x=stap->pop(); stap->push(log(x));}
 else if(ist(text,"e^x")) {x=stap->pop(); stap->push(exp(x));}
 else if(ist(text,"LOG")) {x=stap->pop(); stap->push(log10(x));}
 else if(ist(text," 10^x")) {x=stap->pop(); stap->push(pow(10,x));}
 else if(ist(text,"y^x")) {x=stap->pop(); y=stap->pop(); stap->push(pow(y,x));}
 else if(ist(text,"y^1/x"))
     {x=stap->pop(); y=stap->pop(); stap->push(pow(y,1.0/x));}
 */
 else if(ist(text," 1/x"))
   {x=stap->pop();
    if(x==0) stap->push(1e31);
    else stap->push(1.0/x);
   }
 else if(ist(text,"RAD")) winkelmodus=RADIAN;
 else if(ist(text,"DEG")) winkelmodus=DEGREE;
 else if(ist(text,"STO")) regist[0]=stap->get(0);
 else if(ist(text,"RCL")) stap->push(regist[0]);
 else if(ist(text,"STOx"))
	{x=stap->pop(); c=int(abs(x)+0.001); c2=int(abs(x)+0.999);
	 if(c==c2 && c>=1 && c<=9) regist[c]=stap->get(0);
	 else {stap->push(x); meldung.error("only register 1 to 9 available");}
	}
 else if(ist(text,"RCLx"))
	{x=stap->pop(); c=int(abs(x)+0.001); c2=int(abs(x)+0.999);
	 if(c==c2 && c>=1 && c<=9) stap->push(regist[c]);
	 else {stap->push(x); meldung.error("only register 1 to 9 available");}
	}
 /*
 else if(ist(text,"Help"))
	{meldung.say("press key to get help for ","it"); ef->help=1;}
 */
 else if(ist(text,"HEX"))
   {hexmodus=1;}
 else if(ist(text," Dez"))
   {hexmodus=0;}
 else meldung.error("Err:'%s'",text);
 return blue;
}

void tarech_init()
{
 vektar.init();
 meldung.error("Ready");
}

char tastennummer(uchar c)
{
 char str[6];
 switch(c)
  {case 'a': strcpy(str,"Gold"); break;
   case 'b': strcpy(str,"Blue"); break;
   case 'c': strcpy(str,"Help"); break;
   case 'd': strcpy(str,"POP"); break;
   case 'e': strcpy(str,"CORR"); break;
   case '=': strcpy(str,"ENTR"); break;
   case 'I': strcpy(str,"+/-"); break;
   case 'J': strcpy(str,"x<>y"); break;
   case 'N': strcpy(str,"EXP"); break;
   default: str[0]=c; str[1]=0;
  }
 return gettastennummer(str);
}

void tarech_tastenreaktion(uchar taste)
{
 static uchar gold=0,blue=0;
 char i,j=0;
 i=tastennummer(taste);
 //wprintf(3,"i=%d  ",i);//test
 if(i==0) {gold ^= 1;} //Gold bleibt gesetzt
 if(i==1) {blue ^= 1;} //Blue bleibt nur fuer naechsten Tastendruck
 if(gold) j=2;
 if(blue) j++;
 if(i>1) //Gold oder Bluetaste sind schon ausgwertet
  {vektar.shift(j);
   vektar.press(i);
   blue=0; //Blue ist nur fuer einen Tastendruck gueltig
  }
 if(gold) {j=2; lcd_gowrite(1,19,"G");}
 else          {lcd_gowrite(1,19," ");}
 if(blue) {j++; lcd_gowrite(1,20,"B");}
 else          {lcd_gowrite(1,20," ");}
 if(vektar.eingabeaktiv())
      glcd_goto(4,19); //Cursor hinter letztem eingegebenen Zeichen
 else glcd_goto(4,20); //sonst Cursor auf letzter Zeile ganz rechts
}
